from typing import Any, Dict, List, Optional

import httpx

from featureflags.evaluations.segment import Segment

from ..client import AuthenticatedClient
from ..types import Response


def _get_kwargs(
    *,
    client: AuthenticatedClient,
    environment_uuid: str,
) -> Dict[str, Any]:
    url = "{}/client/env/{environmentUUID}/target-segments".format(
        client.base_url, environmentUUID=environment_uuid
    )

    headers: Dict[str, Any] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    return {
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
    }


def _parse_response(*, response: httpx.Response) -> Optional[List[Segment]]:
    if response.status_code == 200:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = Segment.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200
    return None


def _build_response(*, response: httpx.Response) -> Response[List[Segment]]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient,
    environment_uuid: str,
) -> Response[List[Segment]]:
    kwargs = _get_kwargs(
        client=client,
        environment_uuid=environment_uuid,
    )

    response = httpx.get(
        **kwargs,
    )

    return _build_response(response=response)


def sync(
    *,
    client: AuthenticatedClient,
    environment_uuid: str,
) -> Optional[List[Segment]]:
    """All segment groups in project environment"""

    return sync_detailed(
        client=client,
        environment_uuid=environment_uuid,
    ).parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient,
    environment_uuid: str,
) -> Response[List[Segment]]:
    kwargs = _get_kwargs(
        client=client,
        environment_uuid=environment_uuid,
    )

    async with httpx.AsyncClient() as _client:
        response = await _client.get(**kwargs)

    return _build_response(response=response)


async def asyncio(
    *,
    client: AuthenticatedClient,
    environment_uuid: str,
) -> Optional[List[Segment]]:
    """All segment groups"""

    return (
        await asyncio_detailed(
            client=client,
            environment_uuid=environment_uuid,
        )
    ).parsed



import json

import pytest

from featureflags.evaluations.variation import Variation


@pytest.mark.parametrize(
    "test_input,default,expected",
    [
        ("true", False, True),
        ("false", False, False),
        ("", False, False),
    ],
)
def test_variation_bool(test_input, default, expected):
    variation = Variation(identifier="test", value=test_input)

    got = variation.bool(default=default)

    assert got == expected


@pytest.mark.parametrize(
    "test_input,default,expected",
    [
        ("test", "test_default", "test"),
        (None, "test_default", "test_default"),
    ],
)
def test_variation_string(test_input, default, expected):
    variation = Variation(identifier="test", value=test_input)

    got = variation.string(default=default)

    assert got == expected


@pytest.mark.parametrize(
    "test_input,default,expected", [(1.1, 1.0, 1.1), (None, 1.0, 1.0)]
)
def test_variation_number(test_input, default, expected):
    variation = Variation(identifier="test", value=test_input)

    got = variation.number(default=default)

    assert got == expected


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (json.dumps({"field": "value"}), json.dumps({"field": "value"})),
        (None, json.dumps({"field": "value2"})),
    ],
)
def test_variation_json(test_input, expected):
    variation = Variation(identifier="test", value=test_input)

    got = variation.json(default={"field": "value2"})

    assert got == json.loads(expected)



import json
from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from featureflags.models import UNSET, Unset

T = TypeVar("T", bound="Variation")


@attr.s(auto_attribs=True)
class Variation(object):
    identifier: str
    value: str
    name: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def bool(self, default: bool = False) -> bool:
        if self.value:
            return self.value == "true"
        return default

    def string(self, default: str) -> str:
        return self.value or default

    def number(self, default: float) -> float:
        if self.value:
            return float(self.value)
        return default

    def int(self, default: int) -> int:
        if self.value:
            return int(self.value)
        return default

    def json(self, default: dict) -> dict:
        if self.value:
            return json.loads(self.value)
        return default

    def to_dict(self) -> Dict[str, Any]:
        identifier = self.identifier
        value = self.value
        name = self.name
        description = self.description

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "identifier": identifier,
                "value": value,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name
        if description is not UNSET:
            field_dict["description"] = description

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        identifier = d.pop("identifier")

        value = d.pop("value")

        name = d.pop("name", UNSET)

        description = d.pop("description", UNSET)

        variation = cls(
            identifier=identifier,
            value=value,
            name=name,
            description=description,
        )

        variation.additional_properties = d
        return variation

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import pytest

from featureflags.ftypes.number import Number


@pytest.mark.parametrize(
    ["arg", "input", "mocked", "method", "expected"],
    [
        ("", [], None, "starts_with", False),
        ("", [], None, "ends_with", False),
        ("", [], None, "match", False),
        ("", [], None, "contains", False),
        ("", [], None, "equal_sensitive", False),
        (1.1, 1.1, 1.1, "equal", True),
        (2.0, 1.0, 1.0, "greater_than", True),
        (2.0, 1.0, 2.0, "greater_than_equal", True),
        (1.0, 1.0, 2.0, "less_than", True),
        (1.1, 1.1, 2.0, "less_than_equal", True),
    ],
)
def test_type_methods(mocker, arg, input, mocked, method, expected):

    m = mocker.patch("featureflags.ftypes.number.get_float_value",
                     return_value=mocked)
    number = Number(value=arg)

    fn = getattr(number, method)
    got = fn(input)

    assert got == expected
    if mocked:
        m.assert_called_with(input)


def test_in_list():

    number = Number(value=1)

    got = number.in_list([1, 2])
    expected = True

    assert got == expected




import threading
from threading import Thread
from typing import List, Union

from .api.client import AuthenticatedClient
from .api.default.get_feature_config_by_identifier import \
    sync as get_feature_config
from .api.default.get_segment_by_identifier import sync as get_target_segment
from .config import Config
from .interface import Cache
from .models.message import Message
from .sse_client import SSEClient
from .util import log


class StreamProcessor(Thread):
    def __init__(self, cache: Cache, client: AuthenticatedClient,
                 environment_id: str, api_key: str, token: str,
                 config: Config, ready: threading.Event):

        Thread.__init__(self)
        self.daemon = True
        self._running = False
        self._cache = cache
        self._ready = ready
        self._client = client
        self._environment_id = environment_id
        self._api_key = api_key
        self._token = token
        self._stream_url = f'{config.base_url}/stream'
        self._msg_processors: List[Union[FlagMsgProcessor,
                                         SegmentMsgProcessor]] = []

    def run(self):
        log.info("Starting StreamingProcessor connecting to uri: " +
                 self._stream_url)
        self._running = True

        while self._running:
            try:
                messages = self._connect()
                for msg in messages:
                    if not self._running:
                        break
                    if msg.data:
                        log.info("message received %s", msg.data)
                        message = Message.from_str(msg.data)
                        self.process_message(message)
                    if self._ready.is_set() is False:
                        self._ready.set()
            except Exception as e:
                log.warn("Unexpected error on stream connection: %s", e)

    def _connect(self) -> SSEClient:
        return SSEClient(self._stream_url, headers={
            'Authorization': f'Bearer {self._token}',
            'API-Key': self._api_key
        })

    def process_message(self, msg: Message) -> None:
        processor: Union[FlagMsgProcessor, SegmentMsgProcessor, None] = None
        if msg.domain == "flag":
            log.debug('Starting flag message processor with %s', msg)
            processor = FlagMsgProcessor(cache=self._cache,
                                         client=self._client,
                                         environment_id=self._environment_id,
                                         msg=msg)
        elif msg.domain == "target-segment":
            log.debug('Starting segment message processor with %s', msg)
            processor = SegmentMsgProcessor(
                cache=self._cache,
                client=self._client,
                environment_id=self._environment_id,
                msg=msg
            )
        if processor:
            processor.start()
            self._msg_processors.append(processor)

    def stop(self):
        log.info("Stopping stream processor and msg processors")
        for processor in self._msg_processors:
            processor.stop()
        self._running = False


class FlagMsgProcessor(Thread):

    def __init__(self, cache: Cache, client: AuthenticatedClient,
                 environment_id: str, msg: Message):
        Thread.__init__(self)
        self._cache = cache
        self._client = client
        self._environemnt_id = environment_id
        self._msg = msg

    def run(self):
        log.debug("Fetching flag config '%s' from server",
                  self._msg.identifier)
        fc = get_feature_config(client=self._client,
                                identifier=self._msg.identifier,
                                environment_uuid=self._environemnt_id)
        log.info("Feature config '%s' loaded", fc.feature)
        if self._msg.event == 'create' or self._msg.event == 'patch':
            self._cache.set(f'flags/{fc.feature}', fc)
            log.info('flag %s successfully stored in the cache', fc.feature)
        elif self._msg.event == 'delete':
            self._cache.remove(f'flags/{fc.feature}')
            log.info('flag %s successfully removed from cache', fc.feature)


class SegmentMsgProcessor(Thread):

    def __init__(self, cache: Cache, client: AuthenticatedClient,
                 environment_id: str, msg: Message):
        Thread.__init__(self)
        self._cache = cache
        self._client = client
        self._environemnt_id = environment_id
        self._msg = msg

    def run(self):
        log.debug("Fetching target segment '%s' from server",
                  self._msg.identifier)
        ts = get_target_segment(client=self._client,
                                identifier=self._msg.identifier,
                                environment_uuid=self._environemnt_id)
        log.info("Target segment '%s' loaded", ts.identifier)
        if self._msg.event == 'create' or self._msg.event == 'patch':
            self._cache.set(f'segments/{ts.identifier}', ts)
            log.info('flag %s successfully stored in cache', ts.identifier)
        elif self._msg.event == 'delete':
            self._cache.remove(f'segments/{ts.identifier}')
            log.info('flag %s successfully removed from cache', ts.identifier)



from enum import Enum
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

import attr

from featureflags.models import UNSET, Unset
from featureflags.util import log

from .auth_target import Target
from .constants import SEGMENT_MATCH_OPERATOR
from .enum import FeatureState
from .prerequisite import Prerequisite
from .segment import Segments
from .serve import Serve
from .serving_rule import ServingRule, ServingRules
from .variation import Variation
from .variation_map import VariationMap

T = TypeVar("T", bound="FeatureConfig")


@attr.s(auto_attribs=True)
class Evaluation(object):
    flag: str
    variation: Optional[Variation] = None


class FeatureConfigKind(str, Enum):
    BOOLEAN = "boolean"
    INT = "int"
    NUMBER = "number"
    STRING = "string"
    JSON = "json"

    def __str__(self) -> str:
        return str(self.value)


@attr.s(auto_attribs=True)
class FeatureConfig(object):

    default_serve: Serve
    environment: str
    feature: str
    kind: FeatureConfigKind
    off_variation: str
    project: str
    state: FeatureState
    variations: List[Variation]
    rules: ServingRules = ServingRules()
    segments: Optional[Segments] = None
    prerequisites: Union[Unset, List[Prerequisite]] = UNSET
    variation_to_target_map: Union[Unset, List[VariationMap]] = UNSET
    version: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    cache: Dict[str, Any] = {}

    def evaluate(self, target: Target) -> Evaluation:
        log.debug('Flag kind: %s', self.kind)
        variation = None
        if self.kind == FeatureConfigKind.BOOLEAN:
            variation = self.bool_variation(target)
        return Evaluation(flag=self.feature, variation=variation)

    def bool_variation(self, target: Target) -> Optional[Variation]:
        if self.kind != FeatureConfigKind.BOOLEAN:
            return None
        return self.get_variation(target)

    def int_variation(self, target: Target) -> Optional[Variation]:
        if self.kind != FeatureConfigKind.INT:
            return None
        return self.get_variation(target)

    def number_variation(self, target: Target) -> Optional[Variation]:
        if self.kind != FeatureConfigKind.NUMBER:
            return None
        return self.get_variation(target)

    def string_variation(self, target: Target) -> Optional[Variation]:
        if self.kind != FeatureConfigKind.STRING:
            return None
        return self.get_variation(target)

    def get_variation_name(self, target: Target) -> Optional[str]:
        if self.state == FeatureState.OFF:
            return self.off_variation

        if not isinstance(self.variation_to_target_map, Unset):
            for variation_map in self.variation_to_target_map:
                if not isinstance(variation_map.targets, Unset):
                    for _target in variation_map.targets:
                        if target.identifier == _target.identifier:
                            return variation_map.variation

        return self.rules.get_variation_name(target, self.segments,
                                             self.default_serve)

    def get_variation(self, target: Target) -> Optional[Variation]:
        identifier = self.get_variation_name(target)
        variation = next(
            (val for val in self.variations if val.identifier == identifier),
            None
        )
        return variation

    def get_segment_identifiers(self) -> List[str]:
        if 'segments' in self.cache:
            return self.cache['segments']
        identifiers: List[str] = []
        for rule in self.rules:
            for clause in rule.clauses:
                if clause.op == SEGMENT_MATCH_OPERATOR:
                    for identifier in clause.values:
                        identifiers.append(identifier)
        self.cache['segments'] = identifiers
        return identifiers

    def to_dict(self) -> Dict[str, Any]:
        project = self.project
        environment = self.environment
        feature = self.feature
        state = self.state.value

        kind = self.kind.value

        variations = []
        for variations_item_data in self.variations:
            variations_item = variations_item_data.to_dict()

            variations.append(variations_item)

        default_serve = self.default_serve.to_dict()

        off_variation = self.off_variation
        rules: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.rules, Unset):
            rules = []
            for rules_item_data in self.rules:
                rules_item = rules_item_data.to_dict()

                rules.append(rules_item)

        prerequisites: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.prerequisites, Unset):
            prerequisites = []
            for prerequisites_item_data in self.prerequisites:
                prerequisites_item = prerequisites_item_data.to_dict()

                prerequisites.append(prerequisites_item)

        variation_to_target_map: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.variation_to_target_map, Unset):
            variation_to_target_map = []
            for variation_to_target_map_item_data in \
                    self.variation_to_target_map:
                variation_to_target_map_item = (
                    variation_to_target_map_item_data.to_dict()
                )

                variation_to_target_map.append(variation_to_target_map_item)

        version = self.version

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "project": project,
                "environment": environment,
                "feature": feature,
                "state": state,
                "kind": kind,
                "variations": variations,
                "defaultServe": default_serve,
                "offVariation": off_variation,
            }
        )
        if rules is not UNSET:
            field_dict["rules"] = rules
        if prerequisites is not UNSET:
            field_dict["prerequisites"] = prerequisites
        if variation_to_target_map is not UNSET:
            field_dict["variationToTargetMap"] = variation_to_target_map
        if version is not UNSET:
            field_dict["version"] = version

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        project = d.pop("project")

        environment = d.pop("environment")

        feature = d.pop("feature")

        state = FeatureState(d.pop("state"))

        kind = FeatureConfigKind(d.pop("kind"))

        variations = []
        _variations = d.pop("variations")
        for variations_item_data in _variations:
            variations_item = Variation.from_dict(variations_item_data)

            variations.append(variations_item)

        default_serve = Serve.from_dict(d.pop("defaultServe"))

        off_variation = d.pop("offVariation")

        rules: ServingRules = ServingRules()
        _rules = d.pop("rules", UNSET)
        for rules_item_data in _rules or []:
            rules_item = ServingRule.from_dict(rules_item_data)

            rules.append(rules_item)

        prerequisites = []
        _prerequisites = d.pop("prerequisites", UNSET)
        for prerequisites_item_data in _prerequisites or []:
            prerequisites_item = Prerequisite.from_dict(
                prerequisites_item_data
            )

            prerequisites.append(prerequisites_item)

        variation_to_target_map = []
        _variation_to_target_map = d.pop("variationToTargetMap", UNSET)
        for variation_to_target_map_item_data in \
                _variation_to_target_map or []:
            variation_to_target_map_item = VariationMap.from_dict(
                variation_to_target_map_item_data
            )

            variation_to_target_map.append(variation_to_target_map_item)

        version = d.pop("version", UNSET)

        feature_config = cls(
            project=project,
            environment=environment,
            feature=feature,
            state=state,
            kind=kind,
            variations=variations,
            default_serve=default_serve,
            off_variation=off_variation,
            rules=rules,
            prerequisites=prerequisites,
            variation_to_target_map=variation_to_target_map,
            version=version,
        )

        feature_config.additional_properties = d
        return feature_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict

import httpx

from ..client import AuthenticatedClient
from ..types import Response


def _get_kwargs(
    *,
    client: AuthenticatedClient,
    api_key: str,
) -> Dict[str, Any]:
    url = "{}/stream".format(client.base_url)

    headers: Dict[str, Any] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    headers["api-key"] = api_key

    return {
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
    }


def _build_response(*, response: httpx.Response) -> Response[None]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=None,
    )


def sync_detailed(
    *,
    client: AuthenticatedClient,
    api_key: str,
) -> Response[None]:
    kwargs = _get_kwargs(
        client=client,
        api_key=api_key,
    )

    response = httpx.get(
        **kwargs,
    )

    return _build_response(response=response)


async def asyncio_detailed(
    *,
    client: AuthenticatedClient,
    api_key: str,
) -> Response[None]:
    kwargs = _get_kwargs(
        client=client,
        api_key=api_key,
    )

    async with httpx.AsyncClient() as _client:
        response = await _client.get(**kwargs)

    return _build_response(response=response)



import mmh3

from .constants import ONE_HUNDRED


def get_normalized_number(identifier: str, bucket_by: str) -> int:
    return get_normalized_number_with_normalizer(identifier, bucket_by,
                                                 ONE_HUNDRED)


def get_normalized_number_with_normalizer(
    identifier: str, bucket_by: str, normalizer: int
) -> int:
    value = ":".join([bucket_by, identifier])
    hash = int(mmh3.hash(value))
    return (hash % normalizer) + 1



import abc
import typing


class Cache(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "get")
            and callable(subclass.get)
            and hasattr(subclass, "set")
            and callable(subclass.set)
            and hasattr(subclass, "remove")
            and callable(subclass.remove)
            and hasattr(subclass, "keys")
            and callable(subclass.keys)
            or NotImplemented
        )

    @abc.abstractmethod
    def get(self, key: str) -> typing.Any:
        raise NotImplementedError

    @abc.abstractmethod
    def set(self, key: str, value: typing.Any) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def remove(self, keys: typing.List[str]) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def keys(self) -> typing.List[str]:
        raise NotImplementedError



"""
Tests for Server Sent Event (SSE) client

Based on: https://github.com/btubbs/sseclient
"""

import io
import itertools
import json
from unittest import mock

import pytest
import requests
from requests.cookies import RequestsCookieJar

from featureflags import sse_client
from featureflags.sse_client import Event as E


# Some tests of parsing a single event string
def test_round_trip_parse():
    m1 = E(
        data="hi there\nsexy developer",
        event="salutation",
        id="abcdefg",
        retry=10000
    )

    dumped = m1.dump()
    m2 = E.parse(dumped)
    assert m1.id == m2.id
    assert m1.data == m2.data
    assert m1.retry == m2.retry
    assert m1.event == m2.event


def test_no_colon():
    m = E.parse("data")
    assert m.data == ""


def test_no_space():
    m = E.parse("data:hi")
    assert m.data == "hi"


def test_comment():
    raw = ":this is a comment\ndata: this is some data"
    m = E.parse(raw)
    assert m.data == "this is some data"


def test_retry_is_integer():
    m = E.parse("data: hi\nretry: 4000")
    assert m.retry == 4000


def test_default_event():
    m = E.parse("data: blah")
    assert m.event == "message"


def test_eols():
    for eol in ("\r\n", "\r", "\n"):
        m = E.parse("event: hello%sdata: eol%s" % (eol, eol))
        assert m.event == "hello"
        assert m.data == "eol"


class FakeResponse(object):
    def __init__(self, status_code, content, headers=None, encoding="utf-8"):
        self.status_code = status_code
        self.encoding = encoding
        self.apparent_encoding = "utf-8"
        if not isinstance(content, str):
            content = content.decode("utf-8")
        self.stream = content
        self.headers = headers or None
        self.raw = io.BytesIO(content.encode())

    def raise_for_status(self):
        pass

    def iter_content(self, chunk_size=1024):
        return self.raw


def join_events(*events):
    """
    Given a bunch of Event objects, dump them all to strings and join them
    together.
    """
    return "".join(e.dump() for e in events)


# Tests of parsing a multi event stream
@pytest.mark.parametrize("encoding", ["utf-8", None])
def test_last_id_remembered(monkeypatch, encoding):
    content = "data: message 1\nid: abcdef\n\ndata: message 2\n\n"
    fake_get = mock.Mock(return_value=FakeResponse(200, content,
                                                   encoding=encoding))
    monkeypatch.setattr(requests, "get", fake_get)

    c = sse_client.SSEClient("http://blah.com")
    m1 = next(c)
    m2 = next(c)

    assert m1.id == "abcdef"
    assert m2.id is None
    assert c.last_id == "abcdef"


def test_retry_remembered(monkeypatch):
    content = "data: message 1\nretry: 5000\n\ndata: message 2\n\n"
    fake_get = mock.Mock(return_value=FakeResponse(200, content))
    monkeypatch.setattr(requests, "get", fake_get)

    c = sse_client.SSEClient("http://blah.com")
    m1 = next(c)
    m2 = next(c)
    assert m1.retry == 5000
    assert m2.retry is None
    assert c.retry == 5000


def test_extra_newlines_after_event(monkeypatch):
    content = """event: hello
data: hello1


event: hello
data: hello2

event: hello
data: hello3

"""
    fake_get = mock.Mock(return_value=FakeResponse(200, content))
    monkeypatch.setattr(requests, "get", fake_get)

    c = sse_client.SSEClient("http://blah.com")
    m1 = next(c)
    m2 = next(c)
    m3 = next(c)

    assert m1.event == "hello"
    assert m1.data == "hello1"
    assert m2.data == "hello2"
    assert m2.event == "hello"
    assert m3.data == "hello3"
    assert m3.event == "hello"


@pytest.fixture
def multiple_responses(monkeypatch):
    content = join_events(
        E(data="message 1", id="first", retry="2000", event="blah"),
        E(data="message 2", id="second", retry="4000", event="blerg"),
        E(data="message 3\nhas two lines", id="third"),
    )
    fake_get = mock.Mock(return_value=FakeResponse(200, content))
    monkeypatch.setattr(requests, "get", fake_get)

    yield

    fake_get.assert_called_once_with(
        "http://blah.com",
        headers={"Accept": "text/event-stream", "Cache-Control": "no-cache"},
        stream=True,
    )


def assert_multiple_messages(m1, m2, m3):
    assert m1.data == "message 1"
    assert m1.id == "first"
    assert m1.retry == 2000
    assert m1.event == "blah"

    assert m2.data == "message 2"
    assert m2.id == "second"
    assert m2.retry == 4000
    assert m2.event == "blerg"

    assert m3.data == "message 3\nhas two lines"


@pytest.mark.usefixtures("multiple_responses")
def test_multiple_messages():
    c = sse_client.SSEClient("http://blah.com")
    m1 = next(c)
    m2 = next(c)
    m3 = next(c)

    assert_multiple_messages(m1, m2, m3)

    assert c.retry == m2.retry
    assert c.last_id == m3.id


@pytest.mark.usefixtures("multiple_responses")
def test_simple_iteration():
    c = sse_client.SSEClient("http://blah.com")
    m1, m2, m3 = itertools.islice(c, 3)

    assert_multiple_messages(m1, m2, m3)


def test_client_sends_cookies(mocker):
    s = requests.Session()
    s.cookies = RequestsCookieJar()
    s.cookies["foo"] = "bar"
    m = mocker.patch("featureflags.sse_client.requests.Session.send")
    m.return_value.encoding = "utf-8"
    sse_client.SSEClient("http://blah.com", session=s)
    prepared_request = m.call_args[0][0]
    assert prepared_request.headers["Cookie"] == "foo=bar"


@pytest.fixture
def unicode_multibyte_responses(monkeypatch):
    content = join_events(
        E(
            data="ööööööööööööööööööööööööööööööööööööööööööööööööööööööööö",
            id="first",
            retry="2000",
            event="blah",
        ),
        E(
            data="äääääääääääääääääääääääääääääääääääääääääääääääääääääääää",
            id="second",
            retry="4000",
            event="blerg",
        ),
        E(
            data="üüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüü",
            id="third"
        ),
    )
    fake_get = mock.Mock(return_value=FakeResponse(200, content))
    monkeypatch.setattr(requests, "get", fake_get)

    yield

    fake_get.assert_called_once_with(
        "http://blah.com",
        headers={"Accept": "text/event-stream", "Cache-Control": "no-cache"},
        stream=True,
    )


@pytest.mark.usefixtures("unicode_multibyte_responses")
def test_multiple_unicode_messages():
    c = sse_client.SSEClient("http://blah.com", chunk_size=51)
    assert next(c).data == "öööööööööööööööööööööööööööööööööööööööööööööööö" \
                           "ööööööööö"
    assert next(c).data == "ääääääääääääääääääääääääääääääääääääääääääääääää" \
                           "äääääääää"
    assert next(c).data == "üüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüüü" \
                           "üüüüüüüüü"


def test_event_stream():
    """Check whether event.data can be loaded."""
    limit = 50
    url = "https://stream.wikimedia.org/v2/stream/recentchange"
    source = sse_client.SSEClient(url)
    for n, event in enumerate(source, start=1):
        if event.event != "message" or not event.data:
            continue
        try:
            json.loads(event.data)
        except ValueError as e:
            source.resp.close()
            raise e
        if n == limit:
            break
    assert True



from typing import Any, Dict, List, Optional, Type, TypeVar, Union

import attr

from featureflags.ftypes import OPERATORS
from featureflags.ftypes.interface import Interface
from featureflags.models import UNSET, Unset

from .target_attributes import TargetAttributes

T = TypeVar("T", bound="Target")


@attr.s(auto_attribs=True)
class Target(object):

    identifier: str
    account: str
    org: str
    environment: str
    project: str
    name: str
    anonymous: Union[Unset, bool] = UNSET
    attributes: Union[Unset, TargetAttributes] = UNSET
    created_at: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        identifier = self.identifier
        account = self.account
        org = self.org
        environment = self.environment
        project = self.project
        name = self.name
        anonymous = self.anonymous
        attributes: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.attributes, Unset):
            attributes = self.attributes.to_dict()

        created_at = self.created_at

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "identifier": identifier,
                "account": account,
                "org": org,
                "environment": environment,
                "project": project,
                "name": name,
            }
        )
        if anonymous is not UNSET:
            field_dict["anonymous"] = anonymous
        if attributes is not UNSET:
            field_dict["attributes"] = attributes
        if created_at is not UNSET:
            field_dict["createdAt"] = created_at

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        identifier = d.pop("identifier")

        account = d.pop("account")

        org = d.pop("org")

        environment = d.pop("environment")

        project = d.pop("project")

        name = d.pop("name")

        anonymous = d.pop("anonymous", UNSET)

        _attributes = d.pop("attributes", UNSET)
        attributes: Union[Unset, TargetAttributes]
        if isinstance(_attributes, Unset):
            attributes = UNSET
        else:
            attributes = TargetAttributes.from_dict(_attributes)

        created_at = d.pop("createdAt", UNSET)

        target = cls(
            identifier=identifier,
            account=account,
            org=org,
            environment=environment,
            project=project,
            name=name,
            anonymous=anonymous,
            attributes=attributes,
            created_at=created_at,
        )

        target.additional_properties = d
        return target

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties

    def get_attr_value(self, attribute: str) -> Optional[str]:
        result: Any = getattr(self, attribute, None)
        if isinstance(result, Unset) and not isinstance(self.attributes,
                                                        Unset):
            result = self.attributes.get(attribute, None)
        return result

    def get_operator(self, attribute: str) -> Optional[Interface]:
        value: Optional[str] = self.get_attr_value(attribute)
        for _type, klass in OPERATORS.items():
            if isinstance(value, _type):
                operator = OPERATORS.get(_type, None)
                if operator:
                    return klass(value)
        return None



from featureflags.evaluations.auth_target import Target
from featureflags.evaluations.clause import Clause, Clauses
from featureflags.evaluations.constants import EQUAL_OPERATOR


def test_evaluate_clauses():
    target = Target(
        identifier="john",
        anonymous=False,
        attributes={"email": "john@doe.com"}
    )
    clause_1 = Clause(
        attribute="email",
        id="",
        negate=False,
        op=EQUAL_OPERATOR,
        values=["john@doe.com"],
    )
    clause_2 = Clause(
        attribute="anonymous", id="",
        negate=False,
        op=EQUAL_OPERATOR,
        values=[False]
    )
    clauses = Clauses([clause_1, clause_2])

    got = clauses.evaluate(target, [])

    assert got is True



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from featureflags.models import UNSET, Unset

from .auth_target import Target
from .clause import Clause, Clauses
from .tag import Tag

T = TypeVar("T", bound="Segment")


@attr.s(auto_attribs=True)
class Segment(object):
    identifier: str
    name: str
    environment: Union[Unset, str] = UNSET
    tags: Union[Unset, List[Tag]] = UNSET
    included: Union[Unset, List[str]] = UNSET
    excluded: Union[Unset, List[str]] = UNSET
    rules: Union[Unset, 'Clauses'] = UNSET
    created_at: Union[Unset, int] = UNSET
    modified_at: Union[Unset, int] = UNSET
    version: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def evaluate(self, target: Target) -> bool:
        if not isinstance(self.included, Unset):
            if target.identifier in self.included:
                return True

        if not isinstance(self.excluded, Unset):
            if target.identifier in self.excluded:
                return True

        if not isinstance(self.rules, Unset):
            if self.rules.evaluate(target, None):
                return True
        return False

    def to_dict(self) -> Dict[str, Any]:
        identifier = self.identifier
        name = self.name
        environment = self.environment
        tags: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = []
            for tags_item_data in self.tags:
                tags_item = tags_item_data.to_dict()

                tags.append(tags_item)

        included = self.included
        excluded = self.excluded

        rules: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.rules, Unset):
            rules = []
            for rules_item_data in self.rules:
                rules_item = rules_item_data.to_dict()

                rules.append(rules_item)

        created_at = self.created_at
        modified_at = self.modified_at
        version = self.version

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "identifier": identifier,
                "name": name,
            }
        )
        if environment is not UNSET:
            field_dict["environment"] = environment
        if tags is not UNSET:
            field_dict["tags"] = tags
        if included is not UNSET:
            field_dict["included"] = included
        if excluded is not UNSET:
            field_dict["excluded"] = excluded
        if rules is not UNSET:
            field_dict["rules"] = rules
        if created_at is not UNSET:
            field_dict["createdAt"] = created_at
        if modified_at is not UNSET:
            field_dict["modifiedAt"] = modified_at
        if version is not UNSET:
            field_dict["version"] = version

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        identifier = d.pop("identifier")

        name = d.pop("name")

        environment = d.pop("environment", UNSET)

        tags = []
        _tags = d.pop("tags", UNSET)
        for tags_item_data in _tags or []:
            tags_item = Tag.from_dict(tags_item_data)

            tags.append(tags_item)

        included = d.pop("included", UNSET)
        excluded = d.pop("excluded", UNSET)

        rules: Clauses = Clauses()
        _rules = d.pop("rules", UNSET)
        for rules_item_data in _rules or []:
            rules_item = Clause.from_dict(rules_item_data)

            rules.append(rules_item)

        created_at = d.pop("createdAt", UNSET)

        modified_at = d.pop("modifiedAt", UNSET)

        version = d.pop("version", UNSET)

        segment = cls(
            identifier=identifier,
            name=name,
            environment=environment,
            tags=tags,
            included=included,
            excluded=excluded,
            rules=rules,
            created_at=created_at,
            modified_at=modified_at,
            version=version,
        )

        segment.additional_properties = d
        return segment

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties


class Segments(Dict[str, Segment]):

    def evaluate(self, target: Target) -> bool:
        for _, segment in self.items():
            if not segment.evaluate(target):
                return False
        return True



from typing import Any, Dict, List, Optional, Type, TypeVar

import attr

from featureflags.models import Unset

from .auth_target import Target
from .clause import Clause, Clauses
from .segment import Segments
from .serve import Serve

T = TypeVar("T", bound="ServingRule")


@attr.s(auto_attribs=True)
class ServingRule(object):
    clauses: Clauses
    priority: int
    rule_id: str
    serve: Serve
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        rule_id = self.rule_id
        priority = self.priority
        clauses = []
        for clauses_item_data in self.clauses:
            clauses_item = clauses_item_data.to_dict()

            clauses.append(clauses_item)

        serve = self.serve.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "ruleId": rule_id,
                "priority": priority,
                "clauses": clauses,
                "serve": serve,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        rule_id = d.pop("ruleId")

        priority = d.pop("priority")

        clauses: Clauses = Clauses()
        _clauses = d.pop("clauses")
        for clauses_item_data in _clauses:
            clauses_item = Clause.from_dict(clauses_item_data)

            clauses.append(clauses_item)

        serve = Serve.from_dict(d.pop("serve"))

        serving_rule = cls(
            rule_id=rule_id,
            priority=priority,
            clauses=clauses,
            serve=serve,
        )

        serving_rule.additional_properties = d
        return serving_rule

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties


class ServingRules(List[ServingRule]):
    def get_variation_name(
        self,
        target: Target,
        segments: Optional[Segments] = None,
        default_serve: Optional[Serve] = None,
    ) -> Optional[str]:
        for rule in self:
            if not rule.clauses.evaluate(target, segments):
                continue

            if not isinstance(rule.serve.distribution, Unset):
                return rule.serve.distribution.get_key_name(target)

            if not isinstance(rule.serve.variation, Unset):
                return rule.serve.variation

        if default_serve:
            if not isinstance(default_serve.variation, Unset):
                return default_serve.variation

            if not isinstance(default_serve.distribution, Unset):
                return default_serve.distribution.get_key_name(target)

        return None



# from featureflags.client import CfClient
# from featureflags.config import Config, with_base_url, with_stream_enabled


# def test_client_baseurl_conf():
#     client = CfClient("some key",
#                       with_base_url("localhost"),
#                       with_stream_enabled(True))
#     got = client.config.base_url
#     expect = "localhost"

#     assert got == expect

#     got = client.config.enable_stream
#     expect = True

#     assert got == expect


# def test_client_custom_config():
#     config = Config(enable_stream=True)
#     client = CfClient("some key", *(), config=config)

#     got = client.config.enable_stream
#     expect = True

#     assert got == expect



"""Client for interacting with Harness FF server"""

import threading
from typing import Any, Callable, Dict, Optional

from jwt import decode

from featureflags.analytics import AnalyticsService

from .api.client import AuthenticatedClient, Client
from .api.default.authenticate import AuthenticationRequest
from .api.default.authenticate import sync as authenticate
from .config import Config, default_config
from .evaluations.feature import FeatureConfig
from .evaluations.segment import Segments
from .evaluations.target import Target
from .polling import PollingProcessor
from .streaming import StreamProcessor
from .util import log

VERSION: str = "1.0"


class CfClient(object):
    def __init__(
        self, sdk_key: str, *options: Callable, config: Optional[Config] = None
    ):
        self._client = None
        self._auth_token = None
        self._environment_id = None
        self._sdk_key = sdk_key
        self._config = default_config

        if config:
            self._config = config

        for option in options:
            if callable(option):
                option(self._config)

        log.debug("CfClient initialized")
        self.run()

    def run(self):
        self.authenticate()

        streaming_event = threading.Event()
        polling_event = threading.Event()

        self._polling_processor = PollingProcessor(
            client=self._client,
            config=self._config,
            environment_id=self._environment_id,
            ready=polling_event,
            stream_ready=streaming_event
        )
        self._polling_processor.start()

        if self._config.enable_stream:
            self._stream = StreamProcessor(
                cache=self._config.cache,
                client=self._client,
                environment_id=self._environment_id,
                api_key=self._sdk_key,
                token=self._auth_token,
                config=self._config,
                ready=streaming_event
            )
            self._stream.start()

        if self._config.enable_analytics:
            self._analytics = AnalyticsService(
                config=self._config,
                client=self._client,
                environment=self._environment_id
            )

    def get_environment_id(self):
        return self._environment_id

    def authenticate(self):
        client = Client(
            base_url=self._config.base_url,
            events_url=self._config.events_url
        )
        body = AuthenticationRequest(api_key=self._sdk_key)
        response = authenticate(client=client, json_body=body)
        self._auth_token = response.auth_token

        decoded = decode(self._auth_token, options={
                         "verify_signature": False})
        self._environment_id = decoded["environment"]
        self._client = AuthenticatedClient(
            base_url=self._config.base_url,
            events_url=self._config.events_url,
            token=self._auth_token
        )
        self._client.with_headers({"User-Agent": "PythonSDK/" + VERSION})

    def map_segments_from_cache(self, fc: FeatureConfig) -> None:
        if self._config.cache:
            segments = fc.get_segment_identifiers()
            for identifier in segments:
                segment = self._config.cache.get(f'segments/{identifier}')
                if fc.segments is None:
                    fc.segments = Segments({})
                fc.segments[identifier] = segment

    def _variation(self, fn: str, identifier: str, target: Target,
                   default: Any) -> Any:
        if self._config.cache:
            fc = self._config.cache.get(f'flags/{identifier}')
            if fc:
                self.map_segments_from_cache(fc)
                method = getattr(fc, fn, None)
                if method:
                    variation = method(target)
                    if variation is None:
                        log.debug('No variation found')
                        return default
                    self._analytics.enqueue(target, fc, variation)
                    return variation.bool()
                else:
                    log.error("Wrong method name %s", fn)
        return default

    def bool_variation(self, identifier: str, target: Target,
                       default: bool) -> bool:
        return self._variation('bool_variation', identifier, target, default)

    def int_variation(self, identifier: str, target: Target,
                      default: int) -> int:
        return self._variation('int_variation', identifier, target, default)

    def number_variation(self, identifier: str, target: Target,
                         default: float) -> float:
        return self._variation('number_variation', identifier, target, default)

    def string_variation(self, identifier: str, target: Target,
                         default: str) -> str:
        return self._variation('string_variation', identifier, target, default)

    def json_variation(self, identifier: str, target: Target,
                       default: Dict[str, Any]) -> Dict[str, Any]:
        return self._variation('number_variation', identifier, target, default)

    def close(self):
        log.info('closing sdk client')
        self._polling_processor.stop()
        if self._config.enable_stream:
            self._stream.stop()

        if self._config.enable_analytics:
            self._analytics.close()
        log.info('All threads finished')

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



from .boolean import Boolean
from .integer import Integer
from .json import JSON
from .number import Number
from .string import String

OPERATORS: dict = {
    bool: Boolean,
    str: String,
    int: Integer,
    float: Number,
    dict: JSON
}



from typing import (TYPE_CHECKING, Any, Dict, List, Optional, Type, TypeVar,
                    cast)

import attr

from featureflags.ftypes.interface import Interface
from featureflags.util import log

from .auth_target import Target
from .constants import (CONTAINS_OPERATOR, ENDS_WITH_OPERATOR, EQUAL_OPERATOR,
                        EQUAL_SENSITIVE_OPERATOR, GT_OPERATOR, IN_OPERATOR,
                        SEGMENT_MATCH_OPERATOR, STARTS_WITH_OPERATOR)

if TYPE_CHECKING:
    from .segment import Segments  # noqa

T = TypeVar("T", bound="Clause")


@attr.s(auto_attribs=True)
class Clause(object):
    attribute: str
    id: str
    op: str
    negate: bool = False
    values: List[str] = attr.Factory(list)
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def evaluate(self, target: Target, segments: Optional['Segments'],
                 operator: Optional[Interface]) -> bool:
        if self.op is None or self.op == "":
            return False
        _op = self.op.lower()
        if _op == SEGMENT_MATCH_OPERATOR.lower():
            if segments:
                return segments.evaluate(target)
            return False
        if operator:
            if _op == IN_OPERATOR.lower():
                return operator.in_list(self.values)
            if _op == EQUAL_OPERATOR.lower():
                return operator.equal(self.values)
            if _op == GT_OPERATOR.lower():
                return operator.greater_than(self.values)
            if _op == STARTS_WITH_OPERATOR.lower():
                return operator.starts_with(self.values)
            if _op == ENDS_WITH_OPERATOR.lower():
                return operator.ends_with(self.values)
            if _op == CONTAINS_OPERATOR.lower():
                return operator.contains(self.values)
            if _op == EQUAL_SENSITIVE_OPERATOR.lower():
                return operator.equal_sensitive(self.values)
        # unknown operation
        return False

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        attribute = self.attribute
        op = self.op
        values = self.values

        negate = self.negate

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "attribute": attribute,
                "op": op,
                "values": values,
                "negate": negate,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        attribute = d.pop("attribute")

        op = d.pop("op")

        values = cast(List[str], d.pop("values"))

        negate = d.pop("negate")

        clause = cls(
            id=id,
            attribute=attribute,
            op=op,
            values=values,
            negate=negate,
        )

        clause.additional_properties = d
        return clause

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties


class Clauses(List[Clause]):
    def evaluate(self, target: Target, segments: Optional['Segments']) -> bool:
        for clause in self:
            operator = target.get_operator(clause.attribute)
            if operator is None:
                log.warning("operator not found for clause %s", clause)
            if not clause.evaluate(target, segments, operator):
                return False
        return True



import time
from threading import Thread

from .config import Config
from .util import log


class PersistingProcessor(Thread):

    def __init__(self, config: Config, ready: bool) -> None:
        Thread.__init__(self)
        self.daemon = True
        self._config = config
        self._running = False
        self._ready = ready

    def run(self):
        if not self._running:
            log.info("Starting PersistingProcessor with request interval: %d",
                     self._config.pull_interval)
            self._running = True
            while self._running:
                start_time = time.time()
                try:
                    all_data = self._requester.get_all_data()
                    self._store.init(all_data)
                    if not self._ready.is_set() is True \
                            and self._store.initialized is True:
                        log.info("PersistingProcessor initialized ok")
                        self._ready.set()
                except Exception as e:
                    log.exception(
                        'Error: Exception encountered when updating flags. %s',
                        e
                    )

                elapsed = time.time() - start_time
                if elapsed < self._config.poll_interval:
                    time.sleep(self._config.poll_interval - elapsed)

    def stop(self):
        log.info("Stopping PersistingProcessor")
        self._running = False



import pytest

from featureflags.evaluations.auth_target import Target
from featureflags.ftypes import JSON, Boolean, Integer, Number, String


@pytest.mark.parametrize(
    ["target", "attribute", "expected"],
    [
        (Target(identifier="harness"), "identifier", "harness"),
        (Target(identifier="harness", name="Harness"), "name", "Harness"),
        (
            Target(identifier="harness", name="Harness", anonymous=True),
            "anonymous",
            True,
        ),
        (
            Target(identifier="harness", name="Harness",
                   attributes={"height": 180}),
            "height",
            180,
        ),
    ],
)
def test_get_attr_value(target, attribute, expected):

    got = target.get_attr_value(attribute)

    assert got == expected


@pytest.mark.parametrize(
    ["target", "attribute", "expected"],
    [
        (Target(identifier="harness"), "identifier", String),
        (
            Target(identifier="harness", name="Harness", anonymous=True),
            "anonymous",
            Boolean,
        ),
        (
            Target(identifier="harness", attributes={"height": 180}),
            "height",
            Integer
        ),
        (
            Target(identifier="harness", attributes={"weight": 90.5}),
            "weight",
            Number
        ),
        (
            Target(identifier="harness", attributes={
                "custom": {
                    "key": "value"
                }
            }),
            "custom",
            JSON,
        ),
    ],
)
def test_get_operator(target, attribute, expected):

    got = target.get_operator(attribute)

    assert got.__class__ == expected



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.metrics_data import MetricsData
from ..models.target_data import TargetData
from ..models.unset import UNSET, Unset

T = TypeVar("T", bound="Metrics")


@attr.s(auto_attribs=True)
class Metrics:
    """ """

    target_data: Union[Unset, List[TargetData]] = UNSET
    metrics_data: Union[Unset, List[MetricsData]] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        target_data: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.target_data, Unset):
            target_data = []
            for target_data_item_data in self.target_data:
                target_data_item = target_data_item_data.to_dict()

                target_data.append(target_data_item)

        metrics_data: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.metrics_data, Unset):
            metrics_data = []
            for metrics_data_item_data in self.metrics_data:
                metrics_data_item = metrics_data_item_data.to_dict()

                metrics_data.append(metrics_data_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if target_data is not UNSET:
            field_dict["targetData"] = target_data
        if metrics_data is not UNSET:
            field_dict["metricsData"] = metrics_data

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        target_data = []
        _target_data = d.pop("targetData", UNSET)
        for target_data_item_data in _target_data or []:
            target_data_item = TargetData.from_dict(target_data_item_data)

            target_data.append(target_data_item)

        metrics_data = []
        _metrics_data = d.pop("metricsData", UNSET)
        for metrics_data_item_data in _metrics_data or []:
            metrics_data_item = MetricsData.from_dict(metrics_data_item_data)

            metrics_data.append(metrics_data_item)

        metrics = cls(
            target_data=target_data,
            metrics_data=metrics_data,
        )

        metrics.additional_properties = d
        return metrics

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties




from featureflags.evaluations.auth_target import Target
from featureflags.evaluations.clause import Clause, Clauses
from featureflags.evaluations.constants import SEGMENT_MATCH_OPERATOR
from featureflags.evaluations.distribution import Distribution
from featureflags.evaluations.feature import (Evaluation, FeatureConfig,
                                              FeatureConfigKind, FeatureState)
from featureflags.evaluations.segment import Segment, Segments
from featureflags.evaluations.serve import Serve
from featureflags.evaluations.serving_rule import ServingRule, ServingRules
from featureflags.evaluations.variation import Variation
from featureflags.evaluations.variation_map import VariationMap
from featureflags.evaluations.weighted_variation import WeightedVariation


def test_get_key_name():
    target = Target(identifier="john", attributes={"email": "john@doe.com"})
    wv1 = WeightedVariation("true", 70)
    wv2 = WeightedVariation("false", 30)
    distribution = Distribution("email", [wv1, wv2])

    key = distribution.get_key_name(target)

    assert key == "true"


def test_evaluate_basic():

    target = Target(identifier="john", attributes={"email": "john@doe.com"})

    default_serve = Serve(variation="true")

    variations = [
        Variation(identifier="true", value="true"),
        Variation(identifier="false", value="false"),
    ]

    evaluation = Evaluation("bool-flag", variations[0])

    fc = FeatureConfig(
        feature="bool-flag",
        environment="production",
        default_serve=default_serve,
        kind=FeatureConfigKind.BOOLEAN,
        off_variation="false",
        project="default",
        state=FeatureState.ON,
        variations=variations,
    )

    got = fc.evaluate(target)

    assert got == evaluation


def test_evaluate_target_map():
    target = Target(identifier="user1@example.com")

    default_serve = Serve(variation="true")

    variations = [
        Variation(identifier="true", value="true"),
        Variation(identifier="false", value="false"),
    ]

    evaluation = Evaluation("bool-flag", variations[1])

    variation_to_target_map = VariationMap(variation=variations[1].identifier,
                                           targets=[target])

    fc = FeatureConfig(
        feature="bool-flag",
        environment="production",
        default_serve=default_serve,
        kind=FeatureConfigKind.BOOLEAN,
        off_variation="false",
        project="default",
        state=FeatureState.ON,
        variations=variations,
        variation_to_target_map=[variation_to_target_map]
    )

    got = fc.evaluate(target)

    assert got == evaluation


def test_evaluate_segment_match():
    target = Target(identifier="user1@example.com")
    segment = Segment(
        identifier='beta',
        name='Beta',
        included=[target.identifier]
    )

    default_serve = Serve(variation="true")

    variations = [
        Variation(identifier="true", value="true"),
        Variation(identifier="false", value="false"),
    ]

    evaluation = Evaluation("bool-flag", variations[1])

    clause = Clause(
        attribute='',
        id='',
        op=SEGMENT_MATCH_OPERATOR,
        values=['beta']
    )
    clauses = Clauses([clause])

    serve = Serve(variation="false")

    rule = ServingRule(clauses=clauses, priority=1000, rule_id='', serve=serve)
    rules = ServingRules([rule])

    fc = FeatureConfig(
        feature="bool-flag",
        environment="production",
        default_serve=default_serve,
        kind=FeatureConfigKind.BOOLEAN,
        off_variation="false",
        project="default",
        state=FeatureState.ON,
        variations=variations,
        segments=Segments({
            'beta': segment
        }),
        rules=rules
    )

    got = fc.evaluate(target)

    assert got.variation.identifier == evaluation.variation.identifier



import time
from threading import Lock, Thread
from typing import Dict, List

import attr

from featureflags.models.metrics_data_metrics_type import \
    MetricsDataMetricsType

from .api.client import AuthenticatedClient
from .api.default.post_metrics import sync_detailed as post_metrics
from .config import Config
from .evaluations.feature import FeatureConfig
from .evaluations.target import Target
from .evaluations.variation import Variation
from .models.key_value import KeyValue
from .models.metrics import Metrics
from .models.metrics_data import MetricsData
from .models.target_data import TargetData
from .models.unset import Unset
from .util import log

FF_METRIC_TYPE = 'FFMETRICS'
FEATURE_IDENTIFIER_ATTRIBUTE = 'featureIdentifier'
FEATURE_NAME_ATTRIBUTE = 'featureName'
VARIATION_IDENTIFIER_ATTRIBUTE = 'variationIdentifier'
VARIATION_VALUE_ATTRIBUTE = 'featureValue'
TARGET_ATTRIBUTE = 'target'
SDK_VERSION_ATTRIBUTE = 'SDK_VERSION'
SDK_VERSION = '1.0.0'
SDK_TYPE_ATTRIBUTE = 'SDK_TYPE'
SDK_TYPE = 'server'
SDK_LANGUAGE_ATTRIBUTE = 'SDK_LANGUAGE'
SDK_LANGUAGE = 'python'
GLOBAL_TARGET = 'global'


@attr.s(auto_attribs=True)
class AnalyticsEvent(object):
    target: Target
    feature_config: FeatureConfig
    variation: Variation
    count: int = 0


class AnalyticsService(object):

    def __init__(self, config: Config, client: AuthenticatedClient,
                 environment: str) -> None:
        self._lock = Lock()
        self._config = config
        self._client = client
        self._environment = environment
        self._data: Dict[str, AnalyticsEvent] = {}

        self._running = False
        self._runner = Thread(target=self._sync)
        self._runner.daemon = True
        self._runner.start()

    def enqueue(self, target: Target, feature_config: FeatureConfig,
                variation: Variation):
        event: AnalyticsEvent = AnalyticsEvent(
            target=target,
            feature_config=feature_config,
            variation=variation
        )

        self._lock.acquire()
        try:
            key = self.get_key(event)
            if key in self._data:
                self._data[key].count += 1
            else:
                event.count = 1
                self._data[key] = event
        finally:
            self._lock.release()

    def get_key(self, event: AnalyticsEvent) -> str:
        return '{feature}-{variation}-{value}-{target}'.format(
            feature=event.feature_config.feature,
            variation=event.variation.identifier,
            value=event.variation.value,
            target=GLOBAL_TARGET,
        )

    def _sync(self) -> None:
        if not self._running:
            log.info("Starting AnalyticsService with request interval: %d",
                     self._config.events_sync_interval)
            self._running = True
            while self._running:
                time.sleep(self._config.events_sync_interval)
                self._send_data()

    def _send_data(self) -> None:
        if not bool(self._data):
            log.debug('No metrics data!')
            return
        log.debug('Start sending metrics data')
        self._lock.acquire()
        target_data: List[TargetData] = []
        metrics_data: List[MetricsData] = []
        try:
            for _, event in self._data.items():
                if event.target is not None and not event.target.anonymous:
                    target_attributes: List[KeyValue] = []
                    if not isinstance(event.target.attributes, Unset):
                        for key, value in event.target.attributes:
                            target_attributes.append(KeyValue(key, value))
                    target_name = event.target.identifier
                    if event.target.name:
                        target_name = event.target.name

                    td = TargetData(
                        identifier=event.target.identifier,
                        name=target_name,
                        attributes=target_attributes
                    )
                    target_data.append(td)

                metric_attributes: List[KeyValue] = [
                    KeyValue(FEATURE_IDENTIFIER_ATTRIBUTE,
                             event.feature_config.feature),
                    KeyValue(FEATURE_NAME_ATTRIBUTE,
                             event.feature_config.feature),
                    KeyValue(VARIATION_IDENTIFIER_ATTRIBUTE,
                             event.variation.identifier),
                    KeyValue(VARIATION_VALUE_ATTRIBUTE, event.variation.value),
                    KeyValue(SDK_TYPE_ATTRIBUTE, SDK_TYPE),
                    KeyValue(SDK_LANGUAGE_ATTRIBUTE, SDK_LANGUAGE),
                    KeyValue(SDK_VERSION_ATTRIBUTE, SDK_VERSION),
                    KeyValue(TARGET_ATTRIBUTE, GLOBAL_TARGET)
                ]

                md = MetricsData(
                    timestamp=int(round(time.time() * 1000)),
                    count=event.count,
                    metrics_type=MetricsDataMetricsType.FFMETRICS,
                    attributes=metric_attributes
                )
                metrics_data.append(md)
        finally:
            self._data = {}
            self._lock.release()
        body: Metrics = Metrics(target_data=target_data,
                                metrics_data=metrics_data)
        response = post_metrics(client=self._client,
                                environment=self._environment, json_body=body)
        if 400 >= response.status_code < 500:
            log.error(
                'Error while sending metrics data with status code: %d',
                response.status_code
            )
            return
        log.info('Metrics data sent to server')
        return

    def close(self) -> None:
        self._running = False
        if len(self._data) > 0:
            self._send_data()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
